@@ -40,14 +40,20 @@
 #include <glib/gi18n.h>
 #include <gio/gio.h>
 
+#ifdef HAVE_TELEPATHY_GLIB
+#include <telepathy-glib/account-manager.h>
+#endif
+
 #include <libpanel-util/panel-error.h>
 #include <libpanel-util/panel-glib.h>
 #include <libpanel-util/panel-keyfile.h>
 #include <libpanel-util/panel-launch.h>
+#include <libpanel-util/panel-session-manager.h>
 #include <libpanel-util/panel-show.h>
 
 #include "menu.h"
 #include "panel-action-button.h"
+#include "panel-gconf.h"
 #include "panel-globals.h"
 #include "panel-icon-names.h"
 #include "panel-lockdown.h"
@@ -62,6 +68,7 @@
 #define HOME_NAME_KEY           "/apps/nautilus/desktop/home_icon_name"
 #define COMPUTER_NAME_KEY       "/apps/nautilus/desktop/computer_icon_name"
 #define MAX_ITEMS_OR_SUBMENU    5
+#define MAX_BOOKMARK_ITEMS      100
 
 G_DEFINE_TYPE (PanelPlaceMenuItem, panel_place_menu_item, GTK_TYPE_IMAGE_MENU_ITEM)
 G_DEFINE_TYPE (PanelDesktopMenuItem, panel_desktop_menu_item, GTK_TYPE_IMAGE_MENU_ITEM)
@@ -87,18 +94,101 @@ struct _PanelPlaceMenuItemPrivate {
 	gulong       mount_added_id;
 	gulong       mount_changed_id;
 	gulong       mount_removed_id;
-
-	guint        use_image : 1;
 };
 
 struct _PanelDesktopMenuItemPrivate {
 	GtkWidget   *menu;
 	PanelWidget *panel;
 
-	guint        use_image : 1;
-	guint        append_lock_logout : 1;
+#ifdef HAVE_TELEPATHY_GLIB
+	GList            *presence_items;
+	guint             presence_changed_id;
+	TpAccountManager *account_manager;
+#endif
 };
 
+#ifdef HAVE_TELEPATHY_GLIB
+static void
+panel_menu_item_activate_presence (GtkWidget        *menuitem,
+				   TpAccountManager *account_manager)
+{
+	PanelSessionManagerPresenceType  presence_type;
+	TpConnectionPresenceType         tp_presence_type;
+	const char                      *status;
+	char                            *message;
+
+	presence_type = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (menuitem),
+							    "panel-menu-presence"));
+
+	panel_session_manager_set_presence (panel_session_manager_get (),
+					    presence_type);
+
+	tp_presence_type = tp_account_manager_get_most_available_presence (account_manager,
+									   NULL,
+									   &message);
+
+	if (tp_presence_type == TP_CONNECTION_PRESENCE_TYPE_UNSET ||
+	    tp_presence_type == TP_CONNECTION_PRESENCE_TYPE_OFFLINE ||
+	    tp_presence_type == TP_CONNECTION_PRESENCE_TYPE_UNKNOWN ||
+	    tp_presence_type == TP_CONNECTION_PRESENCE_TYPE_ERROR)
+		goto free_message;
+
+	if (presence_type == PANEL_SESSION_MANAGER_PRESENCE_AVAILABLE) {
+		tp_presence_type = TP_CONNECTION_PRESENCE_TYPE_AVAILABLE;
+		status = "available";
+	} else if (presence_type == PANEL_SESSION_MANAGER_PRESENCE_BUSY) {
+		tp_presence_type = TP_CONNECTION_PRESENCE_TYPE_BUSY;
+		status = "busy";
+	} else
+		goto free_message;
+
+	tp_account_manager_set_all_requested_presences (account_manager,
+							tp_presence_type,
+							status, message);
+
+free_message:
+	g_free (message);
+}
+
+static GtkWidget *
+panel_menu_item_presence_new (TpAccountManager                *account_manager,
+			      PanelSessionManagerPresenceType  presence_type,
+			      const char                      *name,
+			      const char                      *icon,
+			      gboolean                         use_icon)
+{
+	GtkWidget *item;
+
+	if (!account_manager)
+		return NULL;
+
+	item = gtk_check_menu_item_new ();
+	setup_menuitem (item, GTK_ICON_SIZE_INVALID, NULL, name);
+	gtk_check_menu_item_set_draw_as_radio (GTK_CHECK_MENU_ITEM (item), TRUE);
+
+	/* TODO: we need to add an icon at the right of this CheckMenuItem */
+#if 0
+	if (use_icon) {
+		GtkWidget *image;
+		image = gtk_image_new_from_icon_name (icon,
+						      panel_menu_icon_get_size ());
+		gtk_container_add (GTK_CONTAINER (item), image);
+	}
+#endif
+
+	g_object_set_data (G_OBJECT (item), "panel-menu-presence",
+			   GINT_TO_POINTER (presence_type));
+
+	g_signal_connect (item, "activate",
+			  G_CALLBACK (panel_menu_item_activate_presence),
+			  account_manager);
+	g_signal_connect (G_OBJECT (item), "button_press_event",
+			  G_CALLBACK (menu_dummy_button_press_event), NULL);
+
+	return item;
+}
+#endif
+
 static void
 activate_uri_on_screen (const char *uri,
 			GdkScreen  *screen)
@@ -138,12 +228,11 @@ activate_desktop_uri (GtkWidget *menuitem,
 	activate_path (menuitem,
 		       g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP));
 }
- 
-static void
-panel_menu_items_append_from_desktop (GtkWidget *menu,
-				      char      *path,
-				      char      *force_name,
-                                      gboolean   use_icon)
+
+static GtkWidget *
+panel_menu_item_desktop_new (char      *path,
+			     char      *force_name,
+			     gboolean   use_icon)
 {
 	GKeyFile  *key_file;
 	gboolean   loaded;
@@ -193,7 +282,7 @@ panel_menu_items_append_from_desktop (GtkWidget *menu,
 		g_key_file_free (key_file);
 		if (path_freeme)
 			g_free (path_freeme);
-		return;
+		return NULL;
 	}
 
 	/* For Application desktop files, respect TryExec */
@@ -202,7 +291,7 @@ panel_menu_items_append_from_desktop (GtkWidget *menu,
 		g_key_file_free (key_file);
 		if (path_freeme)
 			g_free (path_freeme);
-		return;
+		return NULL;
 	}
 	is_application = (strcmp (type, "Application") == 0);
 	g_free (type);
@@ -223,7 +312,7 @@ panel_menu_items_append_from_desktop (GtkWidget *menu,
 				g_key_file_free (key_file);
 				if (path_freeme)
 					g_free (path_freeme);
-				return;
+				return NULL;
 			}
 
 			g_free (prog);
@@ -250,7 +339,6 @@ panel_menu_items_append_from_desktop (GtkWidget *menu,
 
 	panel_util_set_tooltip_text (item, comment);
 
-	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
 	g_signal_connect_data (item, "activate",
 			       G_CALLBACK (panel_menu_item_activate_desktop_file),
 			       g_strdup (full_path),
@@ -276,16 +364,17 @@ panel_menu_items_append_from_desktop (GtkWidget *menu,
 
 	if (path_freeme)
 		g_free (path_freeme);
+
+	return item;
 }
 
-static void
-panel_menu_items_append_place_item (const char *icon_name,
-				    GIcon      *gicon,
-				    const char *title,
-				    const char *tooltip,
-				    GtkWidget  *menu,
-				    GCallback   callback,
-				    const char *uri)
+static GtkWidget *
+panel_menu_item_uri_new (const char *uri,
+			 const char *icon_name,
+			 GIcon      *gicon,
+			 const char *title,
+			 const char *tooltip,
+			 GCallback   callback)
 {
 	GtkWidget *item;
 	char      *user_data;
@@ -298,8 +387,6 @@ panel_menu_items_append_place_item (const char *icon_name,
 
 	panel_util_set_tooltip_text (item, tooltip);
 
-	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
-
 	user_data = g_strdup (uri);
 	g_signal_connect_data (item, "activate", callback, user_data,
 			       (GClosureNotify) g_free, 0);
@@ -308,16 +395,19 @@ panel_menu_items_append_place_item (const char *icon_name,
 			  G_CALLBACK (menu_dummy_button_press_event), NULL);
 
 	setup_uri_drag (item, uri, icon_name);
+
+	return item;
 }
 
 static GtkWidget *
 panel_menu_items_create_action_item_full (PanelActionButtonType  action_type,
 					  const char            *label,
-					  const char            *tooltip)
+					  const char            *tooltip,
+					  gboolean               create_even_if_disabled)
 {
 	GtkWidget *item;
 
-	if (panel_action_get_is_disabled (action_type))
+	if (!create_even_if_disabled && panel_action_get_is_disabled (action_type))
 		return NULL;
 
 	item = gtk_image_menu_item_new ();
@@ -345,7 +435,66 @@ static GtkWidget *
 panel_menu_items_create_action_item (PanelActionButtonType action_type)
 {
 	return panel_menu_items_create_action_item_full (action_type,
-							 NULL, NULL);
+							 NULL, NULL, FALSE);
+}
+
+#define GDM_FLEXISERVER_COMMAND "gdmflexiserver"
+#define GDM_FLEXISERVER_ARGS    "--startnew"
+
+static void
+panel_menu_item_activate_switch_user (GtkWidget *menuitem,
+				      gpointer   user_data)
+{
+	GdkScreen *screen;
+	GAppInfo  *app_info;
+
+	if (panel_lockdown_get_disable_switch_user_s ())
+		return;
+
+	screen = gtk_widget_get_screen (GTK_WIDGET (menuitem));
+	app_info = g_app_info_create_from_commandline (GDM_FLEXISERVER_COMMAND " " GDM_FLEXISERVER_ARGS,
+						       GDM_FLEXISERVER_COMMAND,
+						       G_APP_INFO_CREATE_NONE,
+						       NULL);
+
+	if (app_info) {
+		GdkAppLaunchContext *launch_context;
+		GdkDisplay          *display;
+
+		display = gdk_screen_get_display (screen);
+		launch_context = gdk_display_get_app_launch_context (display);
+		gdk_app_launch_context_set_screen (launch_context, screen);
+
+		g_app_info_launch (app_info, NULL,
+				   G_APP_LAUNCH_CONTEXT (launch_context),
+				   NULL);
+
+		g_object_unref (launch_context);
+		g_object_unref (app_info);
+	}
+}
+
+static GtkWidget *
+panel_menu_items_create_switch_user (gboolean use_icon)
+{
+	GtkWidget *item;
+
+	if (use_icon) {
+		item = panel_image_menu_item_new ();
+        } else {
+		item = gtk_image_menu_item_new ();
+	}
+
+	setup_menu_item_with_icon (item, panel_menu_icon_get_size (),
+				   NULL, NULL, NULL, _("Switch User"));
+
+	g_signal_connect (item, "activate",
+			  G_CALLBACK (panel_menu_item_activate_switch_user),
+			  NULL);
+	g_signal_connect (G_OBJECT (item), "button_press_event",
+			  G_CALLBACK (menu_dummy_button_press_event), NULL);
+
+	return item;
 }
 
 static void
@@ -358,39 +507,70 @@ panel_place_menu_item_append_gtk_bookmarks (GtkWidget *menu)
 
 	GtkWidget   *add_menu;
 	char        *filename;
-	char        *contents;
-	gchar      **lines;
+	GIOChannel  *io_channel;
 	GHashTable  *table;
 	int          i;
+	GSList      *lines = NULL;
 	GSList      *add_bookmarks, *l;
 	PanelBookmark *bookmark;
 
 	filename = g_build_filename (g_get_home_dir (),
 				     BOOKMARKS_FILENAME, NULL);
 
-	if (!g_file_get_contents (filename, &contents, NULL, NULL)) {
-		g_free (filename);
-		return;
-	}
-
+	io_channel = g_io_channel_new_file (filename, "r", NULL);
 	g_free (filename);
 
-	lines = g_strsplit (contents, "\n", -1);
-	g_free (contents);
+	if (!io_channel)
+		return;
+
+	/* We use a hard limit to avoid having users shooting their
+	 * own feet, and to avoid crashing the system if a misbehaving
+	 * application creates a big bookmars file.
+	 */
+	for (i = 0; i < MAX_BOOKMARK_ITEMS; i++) {
+		char      *contents;
+		gsize      length;
+		gsize      terminator_pos;
+		GIOStatus  status;
+
+		status = g_io_channel_read_line (io_channel, &contents, &length, &terminator_pos, NULL);
+
+		if (status != G_IO_STATUS_NORMAL)
+			break;
+
+		if (length == 0)
+			break;
+
+		/* Clear the line terminator (\n), if any */
+		if (terminator_pos > 0)
+			contents[terminator_pos] = '\0';
+
+		lines = g_slist_prepend (lines, contents);
+	}
+
+	g_io_channel_shutdown (io_channel, FALSE, NULL);
+	g_io_channel_unref (io_channel);
+
+	if (!lines)
+		return;
+
+	lines = g_slist_reverse (lines);
 
 	table = g_hash_table_new (g_str_hash, g_str_equal);
 	add_bookmarks = NULL;
 
-	for (i = 0; lines[i]; i++) {
-		if (lines[i][0] && !g_hash_table_lookup (table, lines[i])) {
+	for (l = lines; l; l = l->next) {
+		char *line = (char*) l->data;
+
+		if (line[0] && !g_hash_table_lookup (table, line)) {
 			GFile    *file;
 			char     *space;
 			char     *label;
 			gboolean  keep;
 
-			g_hash_table_insert (table, lines[i], lines[i]);
+			g_hash_table_insert (table, line, line);
 
-			space = strchr (lines[i], ' ');
+			space = strchr (line, ' ');
 			if (space) {
 				*space = '\0';
 				label = g_strdup (space + 1);
@@ -400,11 +580,11 @@ panel_place_menu_item_append_gtk_bookmarks (GtkWidget *menu)
 
 			keep = FALSE;
 
-			if (g_str_has_prefix (lines[i], "x-nautilus-search:"))
+			if (g_str_has_prefix (line, "x-nautilus-search:"))
 				keep = TRUE;
 
 			if (!keep) {
-				file = g_file_new_for_uri (lines[i]);
+				file = g_file_new_for_uri (line);
 				keep = !g_file_is_native (file) ||
 				       g_file_query_exists (file, NULL);
 				g_object_unref (file);
@@ -417,14 +597,15 @@ panel_place_menu_item_append_gtk_bookmarks (GtkWidget *menu)
 			}
 
 			bookmark = g_malloc (sizeof (PanelBookmark));
-			bookmark->full_uri = g_strdup (lines[i]);
+			bookmark->full_uri = g_strdup (line);
 			bookmark->label = label;
 			add_bookmarks = g_slist_prepend (add_bookmarks, bookmark);
 		}
 	}
 
 	g_hash_table_destroy (table);
-	g_strfreev (lines);
+	g_slist_foreach (lines, (GFunc) g_free, NULL);
+	g_slist_free (lines);
 
 	add_bookmarks = g_slist_reverse (add_bookmarks);
 
@@ -451,6 +632,8 @@ panel_place_menu_item_append_gtk_bookmarks (GtkWidget *menu)
 		char *label;
 		char *icon;
 		GFile *file;
+		GIcon *gicon;
+		GtkWidget *menu_item;
 
 		bookmark = l->data;
 		
@@ -489,15 +672,18 @@ panel_place_menu_item_append_gtk_bookmarks (GtkWidget *menu)
 		if (!icon)
 			icon = g_strdup (PANEL_ICON_FOLDER);
 
+		gicon = g_themed_icon_new_with_default_fallbacks (icon);
+
 		//FIXME: drag and drop will be broken for x-nautilus-search uris
-		panel_menu_items_append_place_item (icon, NULL,
-						    label,
-						    tooltip,
-						    add_menu,
-						    G_CALLBACK (activate_uri),
-						    bookmark->full_uri);
+		menu_item = panel_menu_item_uri_new (bookmark->full_uri,
+						     icon, gicon,
+						     label,
+						     tooltip,
+						     G_CALLBACK (activate_uri));
+		gtk_menu_shell_append (GTK_MENU_SHELL (add_menu), menu_item);
 
 		g_free (icon);
+		g_object_unref (gicon);
 		g_free (tooltip);
 		g_free (label);
 		g_free (bookmark->full_uri);
@@ -550,9 +736,8 @@ panel_menu_item_rescan_drive (GtkWidget *menuitem,
 				menuitem_to_screen (menuitem));
 }
 
-static void
-panel_menu_item_append_drive (GtkWidget *menu,
-			      GDrive    *drive)
+static GtkWidget *
+panel_menu_item_drive_new (GDrive *drive)
 {
 	GtkWidget *item;
 	GIcon     *icon;
@@ -575,8 +760,6 @@ panel_menu_item_append_drive (GtkWidget *menu,
 
 	g_free (title);
 
-	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
-
 	g_signal_connect_data (item, "activate",
 			       G_CALLBACK (panel_menu_item_rescan_drive),
 			       g_object_ref (drive),
@@ -584,6 +767,8 @@ panel_menu_item_append_drive (GtkWidget *menu,
 
 	g_signal_connect (G_OBJECT (item), "button_press_event",
 			  G_CALLBACK (menu_dummy_button_press_event), NULL);
+
+	return item;
 }
 
 typedef struct {
@@ -650,9 +835,8 @@ panel_menu_item_mount_volume (GtkWidget *menuitem,
 			volume_mount_cb, mount_data);
 }
 
-static void
-panel_menu_item_append_volume (GtkWidget *menu,
-			       GVolume   *volume)
+static GtkWidget *
+panel_menu_item_volume_new (GVolume *volume)
 {
 	GtkWidget *item;
 	GIcon     *icon;
@@ -675,8 +859,6 @@ panel_menu_item_append_volume (GtkWidget *menu,
 
 	g_free (title);
 
-	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
-
 	g_signal_connect_data (item, "activate",
 			       G_CALLBACK (panel_menu_item_mount_volume),
 			       g_object_ref (volume),
@@ -684,16 +866,18 @@ panel_menu_item_append_volume (GtkWidget *menu,
 
 	g_signal_connect (G_OBJECT (item), "button_press_event",
 			  G_CALLBACK (menu_dummy_button_press_event), NULL);
+
+	return item;
 }
 
-static void
-panel_menu_item_append_mount (GtkWidget *menu,
-			      GMount    *mount)
+static GtkWidget *
+panel_menu_item_mount_new (GMount *mount)
 {
-	GFile  *root;
-	GIcon  *icon;
-	char   *display_name;
-	char   *activation_uri;
+	GtkWidget *item;
+	GFile     *root;
+	GIcon     *icon;
+	char      *display_name;
+	char      *activation_uri;
 
 	icon = g_mount_get_icon (mount);
 	display_name = g_mount_get_name (mount);
@@ -702,16 +886,16 @@ panel_menu_item_append_mount (GtkWidget *menu,
 	activation_uri = g_file_get_uri (root);
 	g_object_unref (root);
 
-	panel_menu_items_append_place_item (NULL, icon,
-					    display_name,
-					    display_name, //FIXME tooltip
-					    menu,
-					    G_CALLBACK (activate_uri),
-					    activation_uri);
+	item = panel_menu_item_uri_new (activation_uri, NULL, icon,
+					display_name,
+					display_name, //FIXME tooltip
+					G_CALLBACK (activate_uri));
 
 	g_object_unref (icon);
 	g_free (display_name);
 	g_free (activation_uri);
+
+	return item;
 }
 
 typedef enum {
@@ -887,24 +1071,28 @@ panel_place_menu_item_append_local_gio (PanelPlaceMenuItem *place_item,
 	}
 
 	for (sl = items; sl; sl = sl->next) {
+		GtkWidget *menu_item;
 		item = sl->data;
+
 		switch (item->type) {
 		case PANEL_GIO_DRIVE:
-			panel_menu_item_append_drive (add_menu, item->u.drive);
+			menu_item = panel_menu_item_drive_new (item->u.drive);
 			g_object_unref (item->u.drive);
 			break;
 		case PANEL_GIO_VOLUME:
-			panel_menu_item_append_volume (add_menu, item->u.volume);
+			menu_item = panel_menu_item_volume_new (item->u.volume);
 			g_object_unref (item->u.volume);
 			break;
 		case PANEL_GIO_MOUNT:
-			panel_menu_item_append_mount (add_menu, item->u.mount);
+			menu_item = panel_menu_item_mount_new (item->u.mount);
 			g_object_unref (item->u.mount);
 			break;
 		default:
 			g_assert_not_reached ();
 		}
 		g_slice_free (PanelGioItem, item);
+
+		gtk_menu_shell_append (GTK_MENU_SHELL (add_menu), menu_item);
 	}
 
 	g_slist_free (items);
@@ -974,8 +1162,13 @@ panel_place_menu_item_append_remote_gio (PanelPlaceMenuItem *place_item,
 	}
 
 	for (sl = add_mounts; sl; sl = sl->next) {
+		GtkWidget *item;
+
 		mount = sl->data;
-		panel_menu_item_append_mount (add_menu, mount);
+
+		item = panel_menu_item_mount_new (mount);
+		gtk_menu_shell_append (GTK_MENU_SHELL (add_menu), item);
+
 		g_object_unref (mount);
 	}
 
@@ -1001,12 +1194,12 @@ panel_place_menu_item_create_menu (PanelPlaceMenuItem *place_item)
 	name = panel_util_get_label_for_uri (uri);
 	g_object_unref (file);
 	
-	panel_menu_items_append_place_item (PANEL_ICON_HOME, NULL,
-					    name,
-					    _("Open your personal folder"),
-					    places_menu,
-					    G_CALLBACK (activate_home_uri),
-					    uri);
+	item = panel_menu_item_uri_new (uri, PANEL_ICON_HOME, NULL,
+					name,
+					_("Open your personal folder"),
+					G_CALLBACK (activate_home_uri));
+	gtk_menu_shell_append (GTK_MENU_SHELL (places_menu), item);
+
 	g_free (name);
 	g_free (uri);
 
@@ -1017,17 +1210,17 @@ panel_place_menu_item_create_menu (PanelPlaceMenuItem *place_item)
 		uri = g_file_get_uri (file);
 		g_object_unref (file);
 		
-		panel_menu_items_append_place_item (
-				PANEL_ICON_DESKTOP, NULL,
+		item = panel_menu_item_uri_new (
+				/* FIXME: if the dir changes, we'd need to update the drag data since the uri is not the same */
+				uri, PANEL_ICON_DESKTOP, NULL,
 				/* Translators: Desktop is used here as in
 				 * "Desktop Folder" (this is not the Desktop
 				 * environment). */
 				C_("Desktop Folder", "Desktop"),
 				_("Open the contents of your desktop in a folder"),
-				places_menu,
-				G_CALLBACK (activate_desktop_uri),
-				/* FIXME: if the dir changes, we'd need to update the drag data since the uri is not the same */
-				uri);
+				G_CALLBACK (activate_desktop_uri));
+		gtk_menu_shell_append (GTK_MENU_SHELL (places_menu), item);
+
 		g_free (uri);
 	}
 
@@ -1037,20 +1230,29 @@ panel_place_menu_item_create_menu (PanelPlaceMenuItem *place_item)
 	gconf_name = gconf_client_get_string (panel_gconf_get_client (),
 					      COMPUTER_NAME_KEY,
 					      NULL);
-	panel_menu_items_append_from_desktop (places_menu,
-					      "nautilus-computer.desktop",
-					      gconf_name,
-                                              TRUE);
-	if (gconf_name)
-		g_free (gconf_name);
+
+	if (gconf_name == NULL) {
+		gconf_name = g_strdup (_("Computer"));
+	}
+
+	item = panel_menu_item_uri_new ("computer://",
+					PANEL_ICON_COMPUTER, NULL,
+					gconf_name,
+					_("Browse all local and remote disks and folders accessible from this computer"),
+					G_CALLBACK (activate_uri));
+	gtk_menu_shell_append (GTK_MENU_SHELL (places_menu), item);
+	g_free (gconf_name);
 
 	panel_place_menu_item_append_local_gio (place_item, places_menu);
 	add_menu_separator (places_menu);
 
-	panel_menu_items_append_from_desktop (places_menu,
-					      "network-scheme.desktop",
-					      NULL,
-                                              TRUE);
+	item = panel_menu_item_uri_new ("network://",
+					PANEL_ICON_NETWORK, NULL,
+					_("Network"),
+					_("Browse bookmarked and local network locations"),
+					G_CALLBACK (activate_uri));
+	gtk_menu_shell_append (GTK_MENU_SHELL (places_menu), item);
+
 	panel_place_menu_item_append_remote_gio (place_item, places_menu);
 
 	if (panel_is_program_in_path ("nautilus-connect-server")) {
@@ -1062,10 +1264,11 @@ panel_place_menu_item_create_menu (PanelPlaceMenuItem *place_item)
 
 	add_menu_separator (places_menu);
 
-	panel_menu_items_append_from_desktop (places_menu,
-					      "gnome-search-tool.desktop",
-					      NULL,
-                                              FALSE);
+	item = panel_menu_item_desktop_new ("gnome-search-tool.desktop",
+					    NULL,
+					    FALSE);
+	if (item)
+		gtk_menu_shell_append (GTK_MENU_SHELL (places_menu), item);
 
 	panel_recent_append_documents_menu (places_menu,
 					    place_item->priv->recent_manager);
@@ -1134,67 +1337,63 @@ panel_place_menu_item_mounts_changed (GVolumeMonitor *monitor,
 	panel_place_menu_item_recreate_menu (place_menu);
 }
 
-static void
-panel_desktop_menu_item_append_menu (GtkWidget *menu,
-				     gpointer   data)
-{
-	PanelDesktopMenuItem *parent;
-	gboolean              add_separator;
-	GList                *children;
-	GList                *last;
-
-	parent = PANEL_DESKTOP_MENU_ITEM (data);
-
-	add_separator = FALSE;
-	children = gtk_container_get_children (GTK_CONTAINER (menu));
-	last = g_list_last (children);
-
-	if (last != NULL)
-		add_separator = !GTK_IS_SEPARATOR (GTK_WIDGET (last->data));
-
-	g_list_free (children);
-
-	if (add_separator)
-		add_menu_separator (menu);
-
-	panel_menu_items_append_from_desktop (menu, "yelp.desktop", NULL, FALSE);
-	panel_menu_items_append_from_desktop (menu, "gnome-about.desktop", NULL, FALSE);
-
-	if (parent->priv->append_lock_logout)
-		panel_menu_items_append_lock_logout (menu);
-}
-
 static GtkWidget *
-panel_desktop_menu_item_create_menu (PanelDesktopMenuItem *desktop_item)
+panel_desktop_menu_item_create_menu (PanelDesktopMenuItem *desktop_item,
+				     gboolean append_lock_logout)
 {
 	GtkWidget *desktop_menu;
+	GtkWidget *item;
+#ifdef HAVE_TELEPATHY_GLIB
+	gboolean   added;
+#endif
 
-	desktop_menu = create_applications_menu ("settings.menu", NULL);
+	desktop_menu = panel_create_menu ();
 
-	g_object_set_data (G_OBJECT (desktop_menu),
-			   "panel-menu-append-callback",
-			   panel_desktop_menu_item_append_menu);
-	g_object_set_data (G_OBJECT (desktop_menu),
-			   "panel-menu-append-callback-data",
-			   desktop_item);
+#ifdef HAVE_TELEPATHY_GLIB
+	desktop_item->priv->account_manager = tp_account_manager_dup ();
+
+	item = panel_menu_item_presence_new (desktop_item->priv->account_manager,
+					     PANEL_SESSION_MANAGER_PRESENCE_AVAILABLE,
+					     _("Available"),
+					     PANEL_ICON_USER_AVAILABLE, TRUE);
+	if (item) {
+		desktop_item->priv->presence_items = g_list_prepend (desktop_item->priv->presence_items, item);
+		gtk_menu_shell_append (GTK_MENU_SHELL (desktop_menu), item);
+		added = TRUE;
+	}
+
+	item = panel_menu_item_presence_new (desktop_item->priv->account_manager,
+					     PANEL_SESSION_MANAGER_PRESENCE_BUSY,
+					     _("Busy"),
+					     PANEL_ICON_USER_BUSY, TRUE);
+	if (item) {
+		desktop_item->priv->presence_items = g_list_prepend (desktop_item->priv->presence_items, item);
+		gtk_menu_shell_append (GTK_MENU_SHELL (desktop_menu), item);
+		added = TRUE;
+	}
+
+	if (added)
+		add_menu_separator (desktop_menu);
+#endif
+
+	item = panel_menu_item_desktop_new ("gnome-user-accounts-panel.desktop",
+					    _("My Account"), FALSE);
+	if (item)
+		gtk_menu_shell_append (GTK_MENU_SHELL (desktop_menu), item);
+
+	/* Do not force the string like in gnome-shell, but just use the one
+	 * from the .desktop file */
+	item = panel_menu_item_desktop_new ("gnome-control-center.desktop",
+					    NULL, FALSE);
+	if (item)
+		gtk_menu_shell_append (GTK_MENU_SHELL (desktop_menu), item);
+
+	if (append_lock_logout)
+		panel_menu_items_append_lock_logout (desktop_menu);
 
 	return desktop_menu;
 }
 
-static void
-panel_desktop_menu_item_recreate_menu (PanelDesktopMenuItem *desktop_item)
-{
-	if (desktop_item->priv->menu) {
-		gtk_widget_destroy (desktop_item->priv->menu);
-		desktop_item->priv->menu = panel_desktop_menu_item_create_menu (desktop_item);
-		gtk_menu_item_set_submenu (GTK_MENU_ITEM (desktop_item),
-					   desktop_item->priv->menu);
-		panel_applet_menu_set_recurse (GTK_MENU (desktop_item->priv->menu),
-					       "menu_panel",
-					       desktop_item->priv->panel);
-	}
-}
-
 static void
 panel_place_menu_item_finalize (GObject *object)
 {
@@ -1270,9 +1469,20 @@ panel_desktop_menu_item_finalize (GObject *object)
 {
 	PanelDesktopMenuItem *menuitem = (PanelDesktopMenuItem *) object;
 
-	if (menuitem->priv->append_lock_logout)
-		panel_lockdown_notify_remove (G_CALLBACK (panel_desktop_menu_item_recreate_menu),
-					      menuitem);
+#ifdef HAVE_TELEPATHY_GLIB
+	g_list_free (menuitem->priv->presence_items);
+	menuitem->priv->presence_items = NULL;
+
+	if (menuitem->priv->presence_changed_id != 0)
+		g_signal_handler_disconnect (panel_session_manager_get (),
+					     menuitem->priv->presence_changed_id);
+	menuitem->priv->presence_changed_id = 0;
+
+	if (menuitem->priv->account_manager != NULL)
+		g_object_unref (menuitem->priv->account_manager);
+	menuitem->priv->account_manager = NULL;
+#endif
+
 	G_OBJECT_CLASS (panel_desktop_menu_item_parent_class)->finalize (object);
 }
 
@@ -1375,14 +1585,20 @@ static void
 panel_desktop_menu_item_init (PanelDesktopMenuItem *menuitem)
 {
 	menuitem->priv = PANEL_DESKTOP_MENU_ITEM_GET_PRIVATE (menuitem);
+
+#ifdef HAVE_TELEPATHY_GLIB
+	menuitem->priv->presence_items = NULL;
+	menuitem->priv->presence_changed_id = 0;
+	menuitem->priv->account_manager = NULL;
+#endif
 }
 
 static void
 panel_place_menu_item_class_init (PanelPlaceMenuItemClass *klass)
 {
-	GObjectClass *gobject_class = (GObjectClass   *) klass;
+	GObjectClass *gobject_class = (GObjectClass *) klass;
 
-	gobject_class->finalize  = panel_place_menu_item_finalize;
+	gobject_class->finalize = panel_place_menu_item_finalize;
 
 	g_type_class_add_private (klass, sizeof (PanelPlaceMenuItemPrivate));
 }
@@ -1390,9 +1606,9 @@ panel_place_menu_item_class_init (PanelPlaceMenuItemClass *klass)
 static void
 panel_desktop_menu_item_class_init (PanelDesktopMenuItemClass *klass)
 {
-	GObjectClass *gobject_class = (GObjectClass   *) klass;
+	GObjectClass *gobject_class = (GObjectClass *) klass;
 
-	gobject_class->finalize  = panel_desktop_menu_item_finalize;
+	gobject_class->finalize = panel_desktop_menu_item_finalize;
 
 	g_type_class_add_private (klass, sizeof (PanelDesktopMenuItemPrivate));
 }
@@ -1416,8 +1632,6 @@ panel_place_menu_item_new (gboolean use_image)
 			image,
 			_("Places"));
 
-	menuitem->priv->use_image = use_image;
-
 	menuitem->priv->menu = panel_place_menu_item_create_menu (menuitem);
 	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem),
 				   menuitem->priv->menu);
@@ -1425,37 +1639,114 @@ panel_place_menu_item_new (gboolean use_image)
 	return GTK_WIDGET (menuitem);
 }
 
+#ifdef HAVE_TELEPATHY_GLIB
+static void
+panel_desktop_menu_item_on_presence_changed (PanelSessionManager             *manager,
+					     PanelSessionManagerPresenceType  presence_type,
+					     PanelDesktopMenuItem            *desktop_item)
+{
+	const char *icon;
+	GtkWidget  *image;
+	GList      *l;
+
+	switch (presence_type) {
+	case PANEL_SESSION_MANAGER_PRESENCE_AVAILABLE:
+		icon = PANEL_ICON_USER_AVAILABLE;
+		break;
+        case PANEL_SESSION_MANAGER_PRESENCE_INVISIBLE:
+		icon = PANEL_ICON_USER_INVISIBLE;
+		break;
+        case PANEL_SESSION_MANAGER_PRESENCE_BUSY:
+		icon = PANEL_ICON_USER_BUSY;
+		break;
+	case PANEL_SESSION_MANAGER_PRESENCE_IDLE:
+	default:
+		icon = PANEL_ICON_USER_IDLE;
+		break;
+	}
+
+	image = gtk_image_menu_item_get_image (GTK_IMAGE_MENU_ITEM (desktop_item));
+	/* we only have an image if we are specifically using an icon for this
+	 * menu */
+	if (image) {
+		gtk_image_set_from_icon_name (GTK_IMAGE (image),
+					      icon, panel_menu_icon_get_size ());
+	}
+
+	for (l = desktop_item->priv->presence_items; l != NULL; l = l->next) {
+		PanelSessionManagerPresenceType for_presence;
+		GObject *object = l->data;
+
+		for_presence = GPOINTER_TO_INT (g_object_get_data (object,
+								   "panel-menu-presence"));
+		g_signal_handlers_block_by_func (object,
+						 panel_menu_item_activate_presence,
+						 desktop_item->priv->account_manager);
+		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (object),
+						for_presence == presence_type);
+		g_signal_handlers_unblock_by_func (object,
+						   panel_menu_item_activate_presence,
+						   desktop_item->priv->account_manager);
+	}
+}
+#endif
+
 GtkWidget *
 panel_desktop_menu_item_new (gboolean use_image,
 			     gboolean append_lock_logout)
 {
 	PanelDesktopMenuItem *menuitem;
-	GtkWidget            *image;
+	char                 *name;
+#ifdef HAVE_TELEPATHY_GLIB
+	PanelSessionManager  *manager;
+#endif
 
 	menuitem = g_object_new (PANEL_TYPE_DESKTOP_MENU_ITEM, NULL);
 
-	if (use_image)
-		image = gtk_image_new_from_icon_name ("computer",
-						      panel_menu_icon_get_size ());
-	else
-		image = NULL;
+	name = panel_util_get_user_name ();
 
-	setup_menuitem (GTK_WIDGET (menuitem),
-			image ? panel_menu_icon_get_size () : GTK_ICON_SIZE_INVALID,
-			image,
-			_("System"));
+	if (use_image) {
+#ifdef HAVE_TELEPATHY_GLIB
+		setup_menu_item_with_icon (GTK_WIDGET (menuitem),
+					   panel_menu_icon_get_size (),
+					   PANEL_ICON_USER_AVAILABLE,
+					   NULL, NULL,
+					   name);
+		gtk_image_menu_item_set_always_show_image (GTK_IMAGE_MENU_ITEM (menuitem),
+							   TRUE);
+#else
+		setup_menu_item_with_icon (GTK_WIDGET (menuitem),
+					   panel_menu_icon_get_size (),
+					   PANEL_ICON_COMPUTER,
+					   NULL, NULL,
+					   name);
+#endif
+	} else
+		setup_menuitem (GTK_WIDGET (menuitem),
+				GTK_ICON_SIZE_INVALID, NULL,
+				name);
 
-	menuitem->priv->use_image = use_image;
+	g_free (name);
 
-	menuitem->priv->append_lock_logout = append_lock_logout;
-	if (append_lock_logout)
-		panel_lockdown_notify_add (G_CALLBACK (panel_desktop_menu_item_recreate_menu),
-					   menuitem);
-
-	menuitem->priv->menu = panel_desktop_menu_item_create_menu (menuitem);
+	menuitem->priv->menu = panel_desktop_menu_item_create_menu (menuitem,
+								    append_lock_logout);
 	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem),
 				   menuitem->priv->menu);
 
+#ifdef HAVE_TELEPATHY_GLIB
+	manager = panel_session_manager_get ();
+
+	menuitem->priv->presence_changed_id =
+		g_signal_connect (manager,
+				  "presence-changed",
+				  G_CALLBACK (panel_desktop_menu_item_on_presence_changed),
+				  menuitem);
+
+	panel_desktop_menu_item_on_presence_changed (manager,
+						     panel_session_manager_get_presence (manager),
+						     menuitem);
+#endif
+
 	return GTK_WIDGET (menuitem);
 }
 
@@ -1485,89 +1776,100 @@ panel_desktop_menu_item_set_panel (GtkWidget   *item,
 				       "menu_panel", panel);
 }
 
+static void
+panel_menu_items_lock_logout_separator_notified (PanelLockdown *lockdown,
+						 gpointer       user_data)
+{
+	GtkWidget *separator = user_data;
+
+	if (!panel_lockdown_get_disable_lock_screen (lockdown) ||
+	    !panel_lockdown_get_disable_switch_user (lockdown) ||
+	    !panel_lockdown_get_disable_log_out (lockdown))
+		gtk_widget_show (separator);
+	else
+		gtk_widget_hide (separator);
+}
+
 void
 panel_menu_items_append_lock_logout (GtkWidget *menu)
 {
-	gboolean    separator_inserted;
 	GList      *children;
 	GList      *last;
 	GtkWidget  *item;
-	const char *translate;
-	char       *label;
-	char       *tooltip;
 
-	separator_inserted = FALSE;
 	children = gtk_container_get_children (GTK_CONTAINER (menu));
 	last = g_list_last (children);
-	if (last != NULL) {
-		separator_inserted = GTK_IS_SEPARATOR (GTK_WIDGET (last->data));
-	}
+	if (last != NULL &&
+	    GTK_IS_SEPARATOR (last->data))
+		item = GTK_WIDGET (last->data);
+	else
+		item = add_menu_separator (menu);
 	g_list_free (children);
 
-	if (panel_lock_screen_action_available ("lock")) {
-		item = panel_menu_items_create_action_item (PANEL_ACTION_LOCK);
-		if (item != NULL) {
-			if (!separator_inserted) {
-				add_menu_separator (menu);
-				separator_inserted = TRUE;
-			}
+	panel_lockdown_on_notify (panel_lockdown_get (),
+				  NULL,
+				  G_OBJECT (item),
+				  panel_menu_items_lock_logout_separator_notified,
+				  item);
+	panel_menu_items_lock_logout_separator_notified (panel_lockdown_get (),
+							 item);
 
-			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
-		}
+	item = panel_menu_items_create_action_item_full (PANEL_ACTION_LOCK,
+							 NULL, NULL, TRUE);
+	if (item != NULL) {
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+		g_object_bind_property (panel_lockdown_get (),
+					"disable-lock-screen",
+					item,
+					"visible",
+					G_BINDING_SYNC_CREATE|G_BINDING_INVERT_BOOLEAN);
 	}
 
-	if (panel_lockdown_get_disable_log_out ())
-		return;
-	/* Below this, we only have log out/shutdown items */
+	item = panel_menu_items_create_switch_user (FALSE);
 
-	/* Translators: translate "1" (msgctxt: "panel:showusername") to anything
-	 * but "1" if "Log Out %s" doesn't make any sense in your
-	 * language (where %s is a username).
-	 */
-	translate = C_("panel:showusername", "1");
-	if (strcmp (translate, "1") == 0) {
-		const char *user_name;
-
-		user_name = g_get_real_name ();
-		if (!user_name || !user_name [0])
-			user_name = g_get_user_name ();
-
-		/* keep those strings in sync with the ones in
-		 * panel-action-button.c */
-		/* Translators: this string is used ONLY if you translated
-		 * "1" (msgctxt: "panel:showusername") to "1" */
-		label = g_strdup_printf (_("Log Out %s..."),
-					 g_get_user_name ());
-		/* Translators: this string is used ONLY if you translated
-		 * "1" (msgctxt: "panel:showusername") to "1" */
-		tooltip = g_strdup_printf (_("Log out %s of this session to "
-					     "log in as a different user"),
-					   user_name);
-	} else {
-		label   = NULL;
-		tooltip = NULL;
+	if (item != NULL) {
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+		g_object_bind_property (panel_lockdown_get (),
+					"disable-switch-user",
+					item,
+					"visible",
+					G_BINDING_SYNC_CREATE|G_BINDING_INVERT_BOOLEAN);
 	}
 
 	item = panel_menu_items_create_action_item_full (PANEL_ACTION_LOGOUT,
-							 label, tooltip);
-	g_free (label);
-	g_free (tooltip);
+							 NULL, NULL, TRUE);
 
 	if (item != NULL) {
-		if (!separator_inserted) {
-			add_menu_separator (menu);
-			separator_inserted = TRUE;
-		}
-
 		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+		g_object_bind_property (panel_lockdown_get (),
+					"disable-log-out",
+					item,
+					"visible",
+					G_BINDING_SYNC_CREATE|G_BINDING_INVERT_BOOLEAN);
 	}
 
-	item = panel_menu_items_create_action_item (PANEL_ACTION_SHUTDOWN);
-	if (item != NULL) {
-		if (!separator_inserted)
-			add_menu_separator (menu);
+	/* FIXME: should be dynamic */
+	if (panel_session_manager_is_shutdown_available (panel_session_manager_get ())) {
+		item = panel_menu_items_create_action_item_full (PANEL_ACTION_SHUTDOWN,
+								 NULL, NULL, TRUE);
+		if (item != NULL) {
+			GtkWidget *sep;
 
-		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+			sep = add_menu_separator (menu);
+
+			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+			g_object_bind_property (panel_lockdown_get (),
+						"disable-log-out",
+						sep,
+						"visible",
+						G_BINDING_SYNC_CREATE|G_BINDING_INVERT_BOOLEAN);
+			g_object_bind_property (panel_lockdown_get (),
+						"disable-log-out",
+						item,
+						"visible",
+						G_BINDING_SYNC_CREATE|G_BINDING_INVERT_BOOLEAN);
+		}
 	}
 }
 
